"""
Boticord API Wrapper
~~~~~~~~~~~~~~~~~~~
A basic wrapper for the BotiCord API.
:copyright: (c) 2022 Marakarka
:license: MIT, see LICENSE for more details.
"""

__title__ = "boticordpy"
__author__ = "Marakarka"
__license__ = "MIT"
__copyright__ = "Copyright 2022 Marakarka"
__version__ = "2.2.0"

from .client import BoticordClient
from .webhook import Webhook

from .types import *



import typing

from . import types as boticord_types
from .http import HttpClient
from .autopost import AutoPost


class BoticordClient:
    """Represents a client that can be used to interact with the BotiCord API.

    Note:
        Remember that every http method can return http exception.

    Args:
        token (:obj:`str`)
            Your bot's Boticord API Token.
    """

    __slots__ = ("http", "_autopost", "_token")

    http: HttpClient

    def __init__(self, token=None):
        self._token = token
        self._autopost: typing.Optional[AutoPost] = None
        self.http = HttpClient(token)

    async def get_bot_info(self, bot_id: int) -> boticord_types.Bot:
        """Gets information about specified bot.

        Args:
            bot_id (:obj:`int`)
                Id of the bot

        Returns:
            :obj:`~.types.Bot`:
                Bot object.
        """
        response = await self.http.get_bot_info(bot_id)
        return boticord_types.Bot(**response)

    async def get_bot_comments(self, bot_id: int) -> list:
        """Gets list of comments of specified bot.

        Args:
            bot_id (:obj:`int`)
                Id of the bot

        Returns:
            :obj:`list` [ :obj:`~.types.SingleComment` ]:
                List of comments.
        """
        response = await self.http.get_bot_comments(bot_id)
        return [boticord_types.SingleComment(**comment) for comment in response]

    async def post_bot_stats(
        self, servers: int = 0, shards: int = 0, users: int = 0
    ) -> dict:
        """Post Bot's stats.

        Args:
            servers ( :obj:`int` )
                Bot's servers count
            shards ( :obj:`int` )
                Bot's shards count
            users ( :obj:`int` )
                Bot's users count
        Returns:
            :obj:`dict`:
                Boticord API Response status
        """
        response = await self.http.post_bot_stats(
            {"servers": servers, "shards": shards, "users": users}
        )
        return response

    async def get_server_info(self, server_id: int) -> boticord_types.Server:
        """Gets information about specified server.

        Args:
            server_id (:obj:`int`)
                Id of the server

        Returns:
            :obj:`~.types.Server`:
                Server object.
        """
        response = await self.http.get_server_info(server_id)
        return boticord_types.Server(**response)

    async def get_server_comments(self, server_id: int) -> list:
        """Gets list of comments of specified server.

        Args:
            server_id (:obj:`int`)
                Id of the server

        Returns:
            :obj:`list` [ :obj:`~.types.SingleComment` ]:
                List of comments.
        """
        response = await self.http.get_server_comments(server_id)
        return [boticord_types.SingleComment(**comment) for comment in response]

    async def post_server_stats(self, payload: dict) -> dict:
        """Post Server's stats. You must be Boticord-Service bot.

        Args:
            payload (:obj:`dict`)
                Custom data (Use Boticord API docs.)
        Returns:
            :obj:`dict`:
                Boticord API Response.
        """
        response = await self.http.post_server_stats(payload)
        return response

    async def get_user_info(self, user_id: int) -> boticord_types.UserProfile:
        """Gets information about specified user.

        Args:
            user_id (:obj:`int`)
                Id of the user

        Returns:
            :obj:`~.types.UserProfile`:
                User Profile object.
        """
        response = await self.http.get_user_info(user_id)
        return boticord_types.UserProfile(**response)

    async def get_user_comments(self, user_id: int) -> boticord_types.UserComments:
        """Gets comments of specified user.

        Args:
            user_id (:obj:`int`)
                Id of the user

        Returns:
            :obj:`~.types.UserComments`:
                User comments on Bots and Servers pages.
        """
        response = await self.http.get_user_comments(user_id)
        return boticord_types.UserComments(**response)

    async def get_user_bots(self, user_id: int) -> list:
        """Gets list of bots of specified user.

        Args:
            user_id (:obj:`int`)
                Id of the user

        Returns:
            :obj:`list` [ :obj:`~.types.SimpleBot` ]:
                List of simple information about users bots.
        """
        response = await self.http.get_user_bots(user_id)
        return [boticord_types.SimpleBot(**bot) for bot in response]

    async def get_my_shorted_links(self, *, code: str = None):
        """Gets shorted links of an authorized user

        Args:
            code (:obj:`str`)
                Code of shorted link. Could be None.

        Returns:
            Union[:obj:`list` [ :obj:`~.types.ShortedLink` ], :obj:`~types.ShortedLink`]:
                List of shorted links if none else shorted link
        """
        response = await self.http.get_my_shorted_links(code)

        return (
            [boticord_types.ShortedLink(**link) for link in response]
            if code is None
            else boticord_types.ShortedLink(**response[0])
        )

    async def create_shorted_link(self, *, code: str, link: str, domain: boticord_types.LinkDomain = 1):
        """Creates new shorted link

        Args:
            code (:obj:`str`)
                Code of link to short.
            link (:obj:`str`)
                Link to short.
            domain (:obj:`~.types.LinkDomain`)
                Domain to use in shorted link

        Returns:
            :obj:`~types.ShortedLink`:
                Shorted Link
        """
        response = await self.http.create_shorted_link(code, link, domain=domain)

        return boticord_types.ShortedLink(**response)

    async def delete_shorted_link(self, code: str, domain: boticord_types.LinkDomain = 1):
        """Deletes shorted link

        Args:
            code (:obj:`str`)
                Code of link to delete.
            domain (:obj:`~.types.LinkDomain`)
                Domain that is used in shorted link

        Returns:
            :obj:`bool`:
                Is link deleted successfully?
        """
        response = await self.http.delete_shorted_link(code, domain)

        return response.get('ok', False)

    def autopost(self) -> AutoPost:
        """Returns a helper instance for auto-posting.

        Returns:
            :obj:`~.autopost.AutoPost`: An instance of AutoPost.
        """
        if self._autopost is not None:
            return self._autopost

        self._autopost = AutoPost(self)
        return self._autopost



import asyncio
import typing

from . import exceptions as bexc


class AutoPost:
    """
    You can use this class to post stats automatically.

    Args:
        client (:obj:`~.client.BoticordClient`)
            An instance of BoticordClient.
    """

    __slots__ = (
        "client",
        "_interval",
        "_success",
        "_error",
        "_stats",
        "_task",
        "_stopped",
    )

    _success: typing.Any
    _error: typing.Any
    _stats: typing.Any
    _task: typing.Optional["asyncio.Task[None]"]

    def __init__(self, client):
        self.client = client
        self._stopped: bool = False
        self._interval: int = 900
        self._task: typing.Optional["asyncio.Task[None]"] = None

    @property
    def is_running(self) -> bool:
        """
        Is autopost running?
        """
        return self._task is not None and not self._task.done()

    def on_success(self, callback: typing.Any = None):
        """
        Registers success callback.

        .. warning::

            Callback functions must be a **coroutine**. If they aren't, then you might get unexpected
            errors. In order to turn a function into a coroutine they must be ``async def``
            functions.

        This method can be used as a decorator (if you want).
        """

        if callback is not None:
            self._success = callback
            return self

        def inner(func):
            if not asyncio.iscoroutinefunction(func):
                raise TypeError(f"<{func.__qualname__}> must be a coroutine function")

            self._success = callback
            return func

        return inner

    def on_error(self, callback: typing.Any = None):
        """
        Registers error callback.

        .. warning::

            Callback functions must be a **coroutine**. If they aren't, then you might get unexpected
            errors. In order to turn a function into a coroutine they must be ``async def``
            functions.

        The callback function requires to take Exception argument.
        This method can be used as a decorator (if you want).
        """
        if callback is not None:
            self._error = callback
            return self

        def inner(func):
            if not asyncio.iscoroutinefunction(func):
                raise TypeError(f"<{func.__qualname__}> must be a coroutine function")

            self._error = callback
            return func

        return inner

    def init_stats(self, callback: typing.Any = None):
        """
        Registers a function that will return stats. Registered Function Must return dictionary.

        .. warning::

            Callback functions must be a **coroutine**. If they aren't, then you might get unexpected
            errors. In order to turn a function into a coroutine they must be ``async def``
            functions.

        This method can be used as a decorator (if you want).
        """
        if callback is not None:
            self._stats = callback
            return self

        def inner(func):
            if not asyncio.iscoroutinefunction(func):
                raise TypeError(f"<{func.__qualname__}> must be a coroutine function")

            self._stats = callback
            return func

        return inner

    @property
    def interval(self) -> float:
        """The interval between posting stats."""
        return self._interval

    def set_interval(self, seconds: int) -> "AutoPost":
        """
        Sets the interval between posting stats.
        Args:
            seconds (:obj:`int`)
                The interval.
        Raises:
            :obj:`ValueError`
                Boticord recommends not to set interval lower than 900 seconds!
        """
        if seconds < 900:
            raise ValueError(
                "no. Boticord recommends not to set interval lower than 900 seconds!"
            )

        self._interval = seconds
        return self

    async def _internal_loop(self) -> None:
        while True:
            stats = await self._stats()
            try:
                await self.client.http.post_bot_stats(stats)
            except Exception as err:
                on_error = getattr(self, "_error", None)
                if on_error:
                    await on_error(err)
            else:
                on_success = getattr(self, "_success", None)
                if on_success:
                    await on_success()

            if self._stopped:
                return None

            await asyncio.sleep(self._interval)

    def start(self):
        """
        Starts the loop.

        Raises:
            :obj:`~.exceptions.InternalException`
                If there's no callback (for getting stats) provided or the autopost is already running.
        """
        if not hasattr(self, "_stats"):
            raise bexc.InternalException("You must provide stats")

        if self.is_running:
            raise bexc.InternalException(
                "Automatically stats posting is already running"
            )

        task = asyncio.ensure_future(self._internal_loop())
        self._task = task
        return task

    def stop(self) -> None:
        """
        Stops the autopost.
        """
        if not self.is_running:
            return None

        self._stopped = True



class BoticordException(Exception):
    """Base exception class for boticordpy.
    This could be caught to handle any exceptions thrown from this library.
    """


class InternalException(BoticordException):
    """Exception that's thrown when an local operation operation fails.

    Attributes
    ----------
    response:
        The meaning of the exception
    """

    def __init__(self, response):
        super().__init__(response)


class HTTPException(BoticordException):
    """Exception that's thrown when an HTTP request operation fails.

    Attributes
    ----------
    response:
        The response of the failed HTTP request.
    """

    def __init__(self, response):
        self.response = response

        fmt = f"{self.response.reason} (Status code: {self.response.status})"

        super().__init__(fmt)


class Unauthorized(HTTPException):
    """Exception that's thrown when status code 401 occurs."""


class Forbidden(HTTPException):
    """Exception that's thrown when status code 403 occurs."""


class NotFound(HTTPException):
    """Exception that's thrown when status code 404 occurs."""


class ToManyRequests(HTTPException):
    """Exception that's thrown when status code 429 occurs."""


class ServerError(HTTPException):
    """Exception that's thrown when status code 500 or 503 occurs."""



import pathlib

from setuptools import setup, find_packages
import re

HERE = pathlib.Path(__file__).parent

with open('boticordpy/__init__.py') as f:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE).group(1)

if version.endswith(('a', 'b', 'rc')):
    # append version identifier based on commit count
    try:
        import subprocess

        p = subprocess.Popen(['git', 'rev-list', '--count', 'HEAD'],
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = p.communicate()
        if out:
            version += out.decode('utf-8').strip()
        p = subprocess.Popen(['git', 'rev-parse', '--short', 'HEAD'],
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = p.communicate()
        if out:
            version += '+g' + out.decode('utf-8').strip()
    except Exception:
        pass

README = (HERE / "README.md").read_text(encoding="utf8")

setup(
    name="boticordpy",
    project_urls={
        "Documentation": "https://py.boticord.top/en/stable",
        "Issue tracker": "https://github.com/boticord/boticordpy/issues",
    },
    packages=find_packages(),
    version=version,
    python_requires=">= 3.6",
    description="A Python wrapper for BotiCord API",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/boticord/boticordpy",
    author="Marakarka",
    author_email="support@kerdoku.top",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    install_requires=["aiohttp"],
)
